/*-
 * Copyright (c) 2008-2010 Damien Bobillot.
 * All rights reserved.
 *
 * This software was developed for the FreeBSD Project by Damien Bobillot
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

#include <sys/param.h>
#include <sys/jail.h>
#include <sys/sysctl.h>

#include <pwd.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <netinet/in.h>

#define PAM_SM_SESSION

#include <security/pam_appl.h>
#include <security/pam_modules.h>
#include <security/openpam.h>

#define _PAMJAIL_MAX_JAILS 32

static int _pamjail_jailpath_to_id(const char* path) {
	// get the list of jails
	struct xprison xps[_PAMJAIL_MAX_JAILS+1];
	size_t len = sizeof(xps);
	if (sysctlbyname("security.jail.list", &xps, &len, NULL, 0) == -1) {
		openpam_log(PAM_LOG_ERROR, "sysctlbyname(security.jail.list): %m");
		return -1;
	}
	if (len < sizeof(xps[0].pr_version) || xps[0].pr_version != XPRISON_VERSION) {
		openpam_log(PAM_LOG_ERROR, "kernel and userland out of sync");
		return -1;
	}
	
	// find our path
#if XPRISON_VERSION == 3
	char* p = (void*)xps;
	while (len >= sizeof(struct xprison)) {
		struct xprison* xp = (struct xprison*)p;
		if (strcmp(path, xp->pr_path) == 0)
			return xp->pr_id;
		// go to next prison
		size_t xp_size = sizeof(struct xprison) + sizeof(struct in_addr)*xp->pr_ip4s + sizeof(struct in6_addr)*xp->pr_ip6s;
		len -= xp_size;
		p += xp_size;
	}
#elif XPRISON_VERSION == 2
	size_t i;
	size_t jcount = len / sizeof(*xps);
	for (i=0; i<jcount; i++) {
		if (strcmp(path, xps[i].pr_path) == 0)
			return xps[i].pr_id;
	}
#else
#error FreeBSD jail version not supported
#endif
	if (len == 0) {
		openpam_log(PAM_LOG_ERROR, "no jail with root path %s", path);
	} else {
		openpam_log(PAM_LOG_ERROR, "too many jails");
	}
	return -1;
}

PAM_EXTERN int
pam_sm_open_session(pam_handle_t *pamh, int flags __unused,
    int argc __unused, const char *argv[] __unused)
{
	const char *dir, *end, *cwd, *user;
	struct passwd *pwd;
	char buf[PATH_MAX];
	int jid;

	if (pam_get_user(pamh, &user, NULL) != PAM_SUCCESS ||
	    user == NULL || (pwd = getpwnam(user)) == NULL)
		return (PAM_SESSION_ERR);
	if (pwd->pw_uid == 0 && !openpam_get_option(pamh, "also_root"))
		return (PAM_SUCCESS);
	if (pwd->pw_dir == NULL)
		return (PAM_SESSION_ERR);
	if ((end = strstr(pwd->pw_dir, "/./")) != NULL) {
		if (snprintf(buf, sizeof(buf), "%.*s",
		    (int)(end - pwd->pw_dir), pwd->pw_dir) > (int)sizeof(buf)) {
			openpam_log(PAM_LOG_ERROR,
			    "%s's home directory is too long", user);
			return (PAM_SESSION_ERR);
		}
		dir = buf;
		cwd = end + 2;
	} else if ((dir = openpam_get_option(pamh, "dir")) != NULL) {
		if ((cwd = openpam_get_option(pamh, "cwd")) == NULL)
			cwd = "/";
	} else {
		if (openpam_get_option(pamh, "always")) {
			openpam_log(PAM_LOG_ERROR,
			    "%s has no chroot directory", user);
			return (PAM_SESSION_ERR);
		}
		return (PAM_SUCCESS);
	}

	openpam_log(PAM_LOG_DEBUG, "chrooting %s to %s", dir, user);
	
	if ((jid = _pamjail_jailpath_to_id(dir)) == -1) {
		return (PAM_SESSION_ERR);
	}
	
	if (jail_attach(jid) == -1) {
		openpam_log(PAM_LOG_ERROR, "jail_attach(): %m");
		return (PAM_SESSION_ERR);
	}
	if (chdir(cwd) == -1) {
		openpam_log(PAM_LOG_ERROR, "chdir(): %m");
		return (PAM_SESSION_ERR);
	}
	pam_setenv(pamh, "HOME", cwd, 1);
	return (PAM_SUCCESS);
}

PAM_EXTERN int
pam_sm_close_session(pam_handle_t *pamh __unused, int flags __unused,
    int argc __unused, const char *argv[] __unused)
{
	return (PAM_SUCCESS);
}

PAM_MODULE_ENTRY("pam_jail");
